/*
 * kgraft_patch_bsc1108963
 *
 * Fix for CVE-2018-14634, bsc#1108963
 *
 *  Upstream commit:
 *  da029c11e6b1 ("exec: Limit arg stack to at most 75% of _STK_LIM")
 *
 *  SLE12(-SP1) commit:
 *  84cc4c81867eaf8f3ba34a53cc278cbe4f73bc3e
 *
 *  SLE12-SP2 and -SP3 commit:
 *  37a5aa2b6467ad2ac780b659f70d9db18c0604a9 ("Linux 4.4.78")
 *
 *  SLE15 commit:
 *  bebe2f3f24ee949c947a7cd4fc6e2edf6e2fae41 ("Expand 4.12.3")
 *
 *
 *  Copyright (c) 2018 SUSE
 *  Author: Nicolai Stange <nstange@suse.de>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/binfmts.h>
#include <linux/mm.h>
#include <linux/uaccess.h>
#include <linux/highmem.h>
#include <linux/fdtable.h>
#include <linux/slab.h>
#include <linux/security.h>
#include <linux/ptrace.h>
#include <linux/fs.h>
#include <linux/fs_struct.h>
#include <linux/file.h>
#include <linux/syscalls.h>
#include "kgr_patch_bsc1108963.h"
#include "kallsyms_relocs.h"


#if !IS_ENABLED(CONFIG_COMPAT)
#error "Live patch supports only CONFIG_COMPAT=y"
#endif

#if !IS_ENABLED(CONFIG_AUDITSYSCALL)
#error "Live patch supports only CONFIG_AUDITSYSCALL=y"
#endif


#define __KGR_DECLARE_TRACE(name, proto, args, cond, data_proto, data_args)	\
	static struct tracepoint *kgr__tracepoint_##name;		\
	static inline void kgr_trace_##name(proto)			\
	{								\
		if (unlikely(static_key_enabled(&kgr__tracepoint_##name->key))) \
			__DO_TRACE(kgr__tracepoint_##name,		\
				TP_PROTO(data_proto),			\
				TP_ARGS(data_args),			\
				TP_CONDITION(cond),,);			\
	}								\

#define KGR_DECLARE_TRACE(name, proto, args)				\
	__KGR_DECLARE_TRACE(name, PARAMS(proto), PARAMS(args),		\
			cpu_online(raw_smp_processor_id()),		\
			PARAMS(void *__data, proto),			\
			PARAMS(__data, args))

#define KGR_TRACE_EVENT(name, proto, args)			\
	KGR_DECLARE_TRACE(name, PARAMS(proto), PARAMS(args))


/* see include/trace/events/sched.h */
KGR_TRACE_EVENT(sched_process_exec,
	TP_PROTO(struct task_struct *p, pid_t old_pid,
		 struct linux_binprm *bprm),

	TP_ARGS(p, old_pid, bprm)
);


struct kmem_cache **kgr_vm_area_cachep;
static void (*kgr_ptrace_notify)(int exit_code);
static int (*kgr_send_sig)(int sig, struct task_struct *p, int priv);
static int (*kgr_unshare_files)(struct files_struct **displaced);
static int (*kgr_prepare_bprm_creds)(struct linux_binprm *bprm);
static void (*kgr_sched_exec)(void);
static struct mm_struct* (*kgr_mm_alloc)(void);
static int (*kgr_init_new_context)(struct task_struct *tsk,
				   struct mm_struct *mm);
static int (*kgr_insert_vm_struct)(struct mm_struct *mm,
				   struct vm_area_struct *vma);
static void (*kgr_proc_exec_connector)(struct task_struct *task);
static void (*kgr_acct_update_integrals)(struct task_struct *tsk);
static void (*kgr_task_numa_free)(struct task_struct *p);
static void (*kgr_free_bprm)(struct linux_binprm *bprm);
static void (*kgr_put_files_struct)(struct files_struct *files);
static void (*kgr_reset_files_struct)(struct files_struct *files);
static void (*kgr_putname)(struct filename *name);

static struct kgr_kallsyms_reloc kgr_funcs[] = {
	{ "__tracepoint_sched_process_exec",
	  (void *)&kgr__tracepoint_sched_process_exec },
	{ "vm_area_cachep", (void *)&kgr_vm_area_cachep },
	{ "ptrace_notify", (void *)&kgr_ptrace_notify },
	{ "send_sig", (void *)&kgr_send_sig },
	{ "unshare_files", (void *)&kgr_unshare_files },
	{ "prepare_bprm_creds", (void *)&kgr_prepare_bprm_creds },
	{ "sched_exec", (void *)&kgr_sched_exec },
	{ "mm_alloc", (void *)&kgr_mm_alloc },
	{ "init_new_context", (void *)&kgr_init_new_context },
	{ "insert_vm_struct", (void *)&kgr_insert_vm_struct },
	{ "proc_exec_connector", (void *)&kgr_proc_exec_connector },
	{ "acct_update_integrals", (void *)&kgr_acct_update_integrals },
	{ "task_numa_free", (void *)&kgr_task_numa_free },
	{ "free_bprm", (void *)&kgr_free_bprm },
	{ "put_files_struct", (void *)&kgr_put_files_struct },
	{ "reset_files_struct", (void *)&kgr_reset_files_struct },
	{ "putname", (void *)&kgr_putname },
};


/* from include/linux/ptrace.h */
/* resolve reference to unexported ptrace_notify() and send_sig() */
static inline void kgr_ptrace_event(int event, unsigned long message)
{
	if (unlikely(ptrace_event_enabled(current, event))) {
		current->ptrace_message = message;
		kgr_ptrace_notify((event << 8) | SIGTRAP);
	} else if (event == PTRACE_EVENT_EXEC) {
		/* legacy EXEC report via SIGTRAP */
		if ((current->ptrace & (PT_PTRACED|PT_SEIZED)) == PT_PTRACED)
			kgr_send_sig(SIGTRAP, current, 0);
	}
}


/* from fs/exec.c */
/* optimized */
static void kgr_acct_arg_size(struct linux_binprm *bprm, unsigned long pages)
{
	struct mm_struct *mm = current->mm;
	long diff = (long)(pages - bprm->vma_pages);

	if (!mm || !diff)
		return;

	bprm->vma_pages = pages;
	add_mm_counter(mm, MM_ANONPAGES, diff);
}

/* inlined */
static void kgr_put_arg_page(struct page *page)
{
	put_page(page);
}

/* inlined */
static void kgr_free_arg_page(struct linux_binprm *bprm, int i)
{
}

/* inlined */
static void kgr_flush_arg_page(struct linux_binprm *bprm, unsigned long pos,
			       struct page *page)
{
	flush_cache_page(bprm->vma, pos, page_to_pfn(page));
}

/* inlined */
static int kgr__bprm_mm_init(struct linux_binprm *bprm)
{
	int err;
	struct vm_area_struct *vma = NULL;
	struct mm_struct *mm = bprm->mm;

	bprm->vma = vma = kmem_cache_zalloc(*kgr_vm_area_cachep, GFP_KERNEL);
	if (!vma)
		return -ENOMEM;

	down_write(&mm->mmap_sem);
	vma->vm_mm = mm;

	/*
	 * Place the stack at the largest stack address the architecture
	 * supports. Later, we'll move this to an appropriate place. We don't
	 * use STACK_TOP because that can depend on attributes which aren't
	 * configured yet.
	 */
	BUILD_BUG_ON(VM_STACK_FLAGS & VM_STACK_INCOMPLETE_SETUP);
	vma->vm_end = STACK_TOP_MAX;
	vma->vm_start = vma->vm_end - PAGE_SIZE;
	vma->vm_flags = VM_SOFTDIRTY | VM_STACK_FLAGS | VM_STACK_INCOMPLETE_SETUP;
	vma->vm_page_prot = vm_get_page_prot(vma->vm_flags);
	INIT_LIST_HEAD(&vma->anon_vma_chain);

	err = kgr_insert_vm_struct(mm, vma);
	if (err)
		goto err;

	mm->stack_vm = mm->total_vm = 1;
	up_write(&mm->mmap_sem);
	bprm->p = vma->vm_end - sizeof(void *);
	return 0;
err:
	up_write(&mm->mmap_sem);
	bprm->vma = NULL;
	kmem_cache_free(*kgr_vm_area_cachep, vma);
	return err;
}

/* inlined */
static bool kgr_valid_arg_len(struct linux_binprm *bprm, long len)
{
	return len <= MAX_ARG_STRLEN;
}

/* inlined */
static int kgr_bprm_mm_init(struct linux_binprm *bprm)
{
	int err;
	struct mm_struct *mm = NULL;

	bprm->mm = mm = kgr_mm_alloc();
	err = -ENOMEM;
	if (!mm)
		goto err;

	err = kgr_init_new_context(current, mm);
	if (err)
		goto err;

	err = kgr__bprm_mm_init(bprm);
	if (err)
		goto err;

	return 0;

err:
	if (mm) {
		bprm->mm = NULL;
		mmdrop(mm);
	}

	return err;
}

struct user_arg_ptr {
#ifdef CONFIG_COMPAT
	bool is_compat;
#endif
	union {
		const char __user *const __user *native;
#ifdef CONFIG_COMPAT
		const compat_uptr_t __user *compat;
#endif
	} ptr;
};

/* inlined */
static const char __user *kgr_get_user_arg_ptr(struct user_arg_ptr argv, int nr)
{
	const char __user *native;

#ifdef CONFIG_COMPAT
	if (unlikely(argv.is_compat)) {
		compat_uptr_t compat;

		if (get_user(compat, argv.ptr.compat + nr))
			return ERR_PTR(-EFAULT);

		return compat_ptr(compat);
	}
#endif

	if (get_user(native, argv.ptr.native + nr))
		return ERR_PTR(-EFAULT);

	return native;
}

/* inlined */
static int kgr_count(struct user_arg_ptr argv, int max)
{
	int i = 0;

	if (argv.ptr.native != NULL) {
		for (;;) {
			const char __user *p = kgr_get_user_arg_ptr(argv, i);

			if (!p)
				break;

			if (IS_ERR(p))
				return -EFAULT;

			if (i >= max)
				return -E2BIG;
			++i;

			if (fatal_signal_pending(current))
				return -ERESTARTNOHAND;
			cond_resched();
		}
	}
	return i;
}

/* inlined */
static int kgr_check_unsafe_exec(struct linux_binprm *bprm)
{
	struct task_struct *p = current, *t;
	unsigned n_fs;
	int res = 0;

	if (p->ptrace) {
		if (p->ptrace & PT_PTRACE_CAP)
			bprm->unsafe |= LSM_UNSAFE_PTRACE_CAP;
		else
			bprm->unsafe |= LSM_UNSAFE_PTRACE;
	}

	/*
	 * This isn't strictly necessary, but it makes it harder for LSMs to
	 * mess up.
	 */
	if (current->no_new_privs)
		bprm->unsafe |= LSM_UNSAFE_NO_NEW_PRIVS;

	n_fs = 1;
	spin_lock(&p->fs->lock);
	rcu_read_lock();
	for (t = next_thread(p); t != p; t = next_thread(t)) {
		if (t->fs == p->fs)
			n_fs++;
	}
	rcu_read_unlock();

	if (p->fs->users > n_fs) {
		bprm->unsafe |= LSM_UNSAFE_SHARE;
	} else {
		res = -EAGAIN;
		if (!p->fs->in_exec) {
			p->fs->in_exec = 1;
			res = 1;
		}
	}
	spin_unlock(&p->fs->lock);

	return res;
}

/* inlined */
static int kgr_exec_binprm(struct linux_binprm *bprm)
{
	pid_t old_pid, old_vpid;
	int ret;

	/* Need to fetch pid before load_binary changes it */
	old_pid = current->pid;
	rcu_read_lock();
	old_vpid = task_pid_nr_ns(current, task_active_pid_ns(current->parent));
	rcu_read_unlock();

	ret = search_binary_handler(bprm);
	if (ret >= 0) {
		kgr_trace_sched_process_exec(current, old_pid, bprm);
		kgr_ptrace_event(PTRACE_EVENT_EXEC, old_vpid);
		current->did_exec = 1;
		kgr_proc_exec_connector(current);

		if (bprm->file) {
			allow_write_access(bprm->file);
			fput(bprm->file);
			bprm->file = NULL; /* to catch use-after-free */
		}
	}

	return ret;
}



/* patched, inlined */
static struct page *kgr_get_arg_page(struct linux_binprm *bprm,
				     unsigned long pos, int write)
{
	struct page *page;
	int ret;

#ifdef CONFIG_STACK_GROWSUP
		ret = expand_downwards(bprm->vma, pos);
		if (ret < 0)
			return NULL;
	}
#endif
	ret = get_user_pages(current, bprm->mm, pos,
			1, write, 1, &page, NULL);
	if (ret <= 0)
		return NULL;

	if (write) {
		unsigned long size = bprm->vma->vm_end - bprm->vma->vm_start;
		/*
		 * Fix CVE-2018-14634
		 *  -2 lines, +1 line
		 */
		unsigned long ptr_size, limit;

		/*
		 * Since the stack will hold pointers to the strings, we
		 * must account for them as well.
		 *
		 * The size calculation is the entire vma while each arg page is
		 * built, so each time we get here it's calculating how far it
		 * is currently (rather than each call being just the newly
		 * added size from the arg page).  As a result, we need to
		 * always add the entire size of the pointers, so that on the
		 * last call to get_arg_page() we'll actually have the entire
		 * correct size.
		 */
		ptr_size = (bprm->argc + bprm->envc) * sizeof(void *);
		if (ptr_size > ULONG_MAX - size)
			goto fail;
		size += ptr_size;

		kgr_acct_arg_size(bprm, size / PAGE_SIZE);

		/*
		 * We've historically supported up to 32 pages (ARG_MAX)
		 * of argument strings even with small stacks
		 */
		if (size <= ARG_MAX)
			return page;

		/*
		 * Fix CVE-2018-14634
		 *  -3 lines, +5 lines
		 */
		/*
		 * Limit to 1/4 of the max stack size or 3/4 of _STK_LIM
		 * (whichever is smaller) for the argv+env strings.
		 * This ensures that:
		 *  - the remaining binfmt code will not run out of stack space,
		 *  - the program will have a reasonable amount of stack left
		 *    to work from.
		 */
		limit = _STK_LIM / 4 * 3;
		limit = min(limit, rlimit(RLIMIT_STACK) / 4);
		if (size > limit)
			goto fail;
	}

	return page;

fail:
	put_page(page);
	return NULL;
}

/* patched, optimized, calls inlined get_arg_page() */
static int kgr_copy_strings(int argc, struct user_arg_ptr argv,
			    struct linux_binprm *bprm)
{
	struct page *kmapped_page = NULL;
	char *kaddr = NULL;
	unsigned long kpos = 0;
	int ret;

	while (argc-- > 0) {
		const char __user *str;
		int len;
		unsigned long pos;

		ret = -EFAULT;
		str = kgr_get_user_arg_ptr(argv, argc);
		if (IS_ERR(str))
			goto out;

		len = strnlen_user(str, MAX_ARG_STRLEN);
		if (!len)
			goto out;

		ret = -E2BIG;
		if (!kgr_valid_arg_len(bprm, len))
			goto out;

		/* We're going to work our way backwords. */
		pos = bprm->p;
		str += len;
		bprm->p -= len;

		while (len > 0) {
			int offset, bytes_to_copy;

			if (fatal_signal_pending(current)) {
				ret = -ERESTARTNOHAND;
				goto out;
			}
			cond_resched();

			offset = pos % PAGE_SIZE;
			if (offset == 0)
				offset = PAGE_SIZE;

			bytes_to_copy = offset;
			if (bytes_to_copy > len)
				bytes_to_copy = len;

			offset -= bytes_to_copy;
			pos -= bytes_to_copy;
			str -= bytes_to_copy;
			len -= bytes_to_copy;

			if (!kmapped_page || kpos != (pos & PAGE_MASK)) {
				struct page *page;

				page = kgr_get_arg_page(bprm, pos, 1);
				if (!page) {
					ret = -E2BIG;
					goto out;
				}

				if (kmapped_page) {
					flush_kernel_dcache_page(kmapped_page);
					kunmap(kmapped_page);
					kgr_put_arg_page(kmapped_page);
				}
				kmapped_page = page;
				kaddr = kmap(kmapped_page);
				kpos = pos & PAGE_MASK;
				kgr_flush_arg_page(bprm, kpos, kmapped_page);
			}
			if (copy_from_user(kaddr+offset, str, bytes_to_copy)) {
				ret = -EFAULT;
				goto out;
			}
		}
	}
	ret = 0;
out:
	if (kmapped_page) {
		flush_kernel_dcache_page(kmapped_page);
		kunmap(kmapped_page);
		kgr_put_arg_page(kmapped_page);
	}
	return ret;
}

/* patched, calls optimized copy_strings() */
int kgr_copy_strings_kernel(int argc, const char *const *__argv,
			    struct linux_binprm *bprm)
{
	int r;
	mm_segment_t oldfs = get_fs();
	struct user_arg_ptr argv = {
		.ptr.native = (const char __user *const  __user *)__argv,
	};

	set_fs(KERNEL_DS);
	r = kgr_copy_strings(argc, argv, bprm);
	set_fs(oldfs);

	return r;
}

/* patched, calls inlined get_arg_page() */
int kgr_remove_arg_zero(struct linux_binprm *bprm)
{
	int ret = 0;
	unsigned long offset;
	char *kaddr;
	struct page *page;

	if (!bprm->argc)
		return 0;

	do {
		offset = bprm->p & ~PAGE_MASK;
		page = kgr_get_arg_page(bprm, bprm->p, 0);
		if (!page) {
			ret = -EFAULT;
			goto out;
		}
		kaddr = kmap_atomic(page);

		for (; offset < PAGE_SIZE && kaddr[offset];
				offset++, bprm->p++)
			;

		kunmap_atomic(kaddr);
		kgr_put_arg_page(page);

		if (offset == PAGE_SIZE)
			kgr_free_arg_page(bprm, (bprm->p >> PAGE_SHIFT) - 1);
	} while (offset == PAGE_SIZE);

	bprm->p++;
	bprm->argc--;
	ret = 0;

out:
	return ret;
}

/* patched, optimized, calls optimized copy_strings() */
static int kgr_do_execve_common(const char *filename,
				struct user_arg_ptr argv,
				struct user_arg_ptr envp)
{
	struct linux_binprm *bprm;
	struct file *file;
	struct files_struct *displaced;
	bool clear_in_exec;
	int retval;

	/*
	 * We move the actual failure in case of RLIMIT_NPROC excess from
	 * set*uid() to execve() because too many poorly written programs
	 * don't check setuid() return code.  Here we additionally recheck
	 * whether NPROC limit is still exceeded.
	 */
	if ((current->flags & PF_NPROC_EXCEEDED) &&
	    atomic_read(&current_user()->processes) > rlimit(RLIMIT_NPROC)) {
		retval = -EAGAIN;
		goto out_ret;
	}

	/* We're below the limit (still or again), so we don't want to make
	 * further execve() calls fail. */
	current->flags &= ~PF_NPROC_EXCEEDED;

	retval = kgr_unshare_files(&displaced);
	if (retval)
		goto out_ret;

	retval = -ENOMEM;
	bprm = kzalloc(sizeof(*bprm), GFP_KERNEL);
	if (!bprm)
		goto out_files;

	retval = kgr_prepare_bprm_creds(bprm);
	if (retval)
		goto out_free;

	retval = kgr_check_unsafe_exec(bprm);
	if (retval < 0)
		goto out_free;
	clear_in_exec = retval;
	current->in_execve = 1;

	file = open_exec(filename);
	retval = PTR_ERR(file);
	if (IS_ERR(file))
		goto out_unmark;

	kgr_sched_exec();

	bprm->file = file;
	bprm->filename = filename;
	bprm->interp = filename;

	retval = kgr_bprm_mm_init(bprm);
	if (retval)
		goto out_file;

	bprm->argc = kgr_count(argv, MAX_ARG_STRINGS);
	if ((retval = bprm->argc) < 0)
		goto out;

	bprm->envc = kgr_count(envp, MAX_ARG_STRINGS);
	if ((retval = bprm->envc) < 0)
		goto out;

	retval = prepare_binprm(bprm);
	if (retval < 0)
		goto out;

	retval = kgr_copy_strings_kernel(1, &bprm->filename, bprm);
	if (retval < 0)
		goto out;

	bprm->exec = bprm->p;
	retval = kgr_copy_strings(bprm->envc, envp, bprm);
	if (retval < 0)
		goto out;

	retval = kgr_copy_strings(bprm->argc, argv, bprm);
	if (retval < 0)
		goto out;

	retval = kgr_exec_binprm(bprm);
	if (retval < 0)
		goto out;

	/* execve succeeded */
	current->fs->in_exec = 0;
	current->in_execve = 0;
	kgr_acct_update_integrals(current);
	kgr_task_numa_free(current);
	kgr_free_bprm(bprm);
	if (displaced)
		kgr_put_files_struct(displaced);
	return retval;

out:
	if (bprm->mm) {
		kgr_acct_arg_size(bprm, 0);
		mmput(bprm->mm);
	}

out_file:
	if (bprm->file) {
		allow_write_access(bprm->file);
		fput(bprm->file);
	}

out_unmark:
	if (clear_in_exec)
		current->fs->in_exec = 0;
	current->in_execve = 0;

out_free:
	kgr_free_bprm(bprm);

out_files:
	if (displaced)
		kgr_reset_files_struct(displaced);
out_ret:
	return retval;
}

/* patched, sometimes inlined, calls optimized do_execve_common() */
int kgr_do_execve(const char *filename,
		  const char __user *const __user *__argv,
		  const char __user *const __user *__envp)
{
	struct user_arg_ptr argv = { .ptr.native = __argv };
	struct user_arg_ptr envp = { .ptr.native = __envp };
	return kgr_do_execve_common(filename, argv, envp);
}

/* patched, inlined, calls optimized do_execve_common() */
static int kgr_compat_do_execve(const char *filename,
				const compat_uptr_t __user *__argv,
				const compat_uptr_t __user *__envp)
{
	struct user_arg_ptr argv = {
		.is_compat = true,
		.ptr.compat = __argv,
	};
	struct user_arg_ptr envp = {
		.is_compat = true,
		.ptr.compat = __envp,
	};
	return kgr_do_execve_common(filename, argv, envp);
}

/* patched, calls inlined do_exceve() */
__SYSCALL_DEFINEx(3, _kgr_execve,
		const char __user *, filename,
		const char __user *const __user *, argv,
		const char __user *const __user *, envp)
{
	struct filename *path = getname(filename);
	int error = PTR_ERR(path);
	if (!IS_ERR(path)) {
		error = kgr_do_execve(path->name, argv, envp);
		kgr_putname(path);
	}
	return error;
}
/* alias for usedsymbols */
asmlinkage long kgr_sys_execve(const char __user *filename,
			const char __user *const __user *argv,
			const char __user *const __user *envp)
	__attribute__((alias("SyS_kgr_execve")));

/* patched, calls inlined compat_do_execve() */
asmlinkage long kgr_compat_sys_execve(const char __user * filename,
				      const compat_uptr_t __user * argv,
				      const compat_uptr_t __user * envp)
{
	struct filename *path = getname(filename);
	int error = PTR_ERR(path);
	if (!IS_ERR(path)) {
		error = kgr_compat_do_execve(path->name, argv, envp);
		kgr_putname(path);
	}
	return error;
}



int kgr_patch_bsc1108963_init(void)
{
	return __kgr_resolve_kallsyms_relocs(kgr_funcs, ARRAY_SIZE(kgr_funcs));
}
